/*
  controller for shift-register based output port

  DAn Wiliams July 26 2004
*/

#include <stdio.h>
#include "parallel.h"
#include "cmnCmdOptions.h"

// -------------- PRE DECLERATIONS ----------------

int  SetQuiet   ( char * unused);
int  DoHelp     ( char * option);
int  ChangePort ( char * option);
int  AllocBits  ( long size );
void SendBit    ( int value);
void Strobe     ( void);
void BitOn      ( long Bit) ;
void BitOff     ( long Bit) ;  
int  CBitOn     ( char * option) ;
int  CBitOff    ( char * option) ;
void AddOptions ( ) ;
int  ChangeBits ( char * option) ;

// --------------- GLOBALS -------------------------


FILE        * defaults;
parallel_t  * ioport;
unsigned int  bits;
char        * BitList;
int           port; 
int           quiet;
int           justexit;


//--------------- CODE ---------------------------

int main (int argc, char **argv) {   
  int   value;
  long  i;
  char  * tempPtr;
  long  newbits;
  
  port     = 0x378;  // default
  bits     = 0;     // default  
  BitList  = NULL;
  defaults = NULL;
  quiet    = 0;
  justexit = 0;
  
  // allocate memory for bits pattern default 16
  AllocBits(16); 
  
  // init command line option setup
  cmdLineOptionInit();

  // try to open defaults file  
  if ((defaults = fopen("/etc/superport.conf", "r")) == NULL) {
    printf("Unable to open defaults file /etc/superport.conf for read\n");
  } else {
      
      // read parallel port to be used
      fscanf(defaults, "%x", &port);
      
      // read number of device bits (multiples of 16)
      fscanf(defaults, "%d", &newbits);
      AllocBits(newbits);
      
      //read default values for 
      for (i = 0; i < bits; i++) {
        fscanf(defaults, "%d", &value);
        if ( value == 0 ) {    
          BitOff(i);
        } else {
          BitOn(i);
        }
      } 
  }
    
  // override with command line options
  AddOptions();
  parseOptions( argc, argv );
  
  if (!quiet) printf(" --- Shift Register Port Controller, Ver 1.2 ---- \n");
  
  if (justexit == 1) {
    if (defaults) {
      fclose(defaults);
    }
    return 0; 
  }
  
  if (!quiet) printf("Device bits: %d \n", bits);
  if (!quiet) printf("Using port: %X \n", port); 
      
  // open parallel port
  ioport = parallel_Init (port);
   
  if (!quiet) printf("Sending: ");   
  for (i = (bits - 1); i >= 0; i-- ) {
    if (BitList[i] ) {
      SendBit(1);
      if (!quiet) printf("1 ");
    } else {
      SendBit(0);  
      if (!quiet) printf("0 ");
    }
  }  
  Strobe();
  if (!quiet) printf(".\n");
  
  if (defaults) {
    fclose(defaults);
  }
 
  if ((defaults = fopen("/etc/superport.conf", "w")) == NULL) {
      printf("Unable to open defaults file /etc/superport.conf for write\n");
  } else {
      fprintf(defaults, "%X\n", port);  // write port number
      fprintf(defaults, "%d\n", bits);  // write machine bits
      for (i = 0; i < bits; i++ ) {     // write out the status of all the bits
        if (BitList[i] ) {
          fprintf(defaults, "1\n");
        } else {
          SendBit(0);  
          fprintf(defaults, "0\n");
        }
      }  
      
      fclose(defaults);
  }

  parallel_Destroy(ioport);
  free(BitList);
  
  return 0;
}


/* ------------------------ FUNCTIONS ----------------------- */


int AllocBits( long newsize ) {
  char * tempPtr;
  long i;
  
  if ((tempPtr = (char *)realloc(BitList, sizeof(char) * newsize)) == NULL) {
    printf("Memory allocation failure, could not allocate %d bytes\n", newsize);
    return 0;     // if fail return 0
  } else {
    BitList = tempPtr; 
    if (newsize > bits) {
      memset(tempPtr+bits, 0, newsize-bits);
    }
    bits = newsize;
    return 1;
  }  
}


void AddOptions() {
    
    static char* offbit[2]    =   {"-0", "--off"};
    static char* onbit[2]     =   {"-1", "--on"};
    static char* bitopts[2]   =   {"--size", "-s"};
    static char* portopts[2]  =   {"--port", "-p"};
    static char* quietopts[2] =   {"--shutup", "-q"};
    static char* helpopts[4]  =   {"--help", "-?", "-h", "?"};

    
    addCmdLineOption(offbit,    2,  CBitOff);
    addCmdLineOption(onbit,     2,  CBitOn);  
    addCmdLineOption(bitopts,   2,  ChangeBits);  
    addCmdLineOption(portopts,  2,  ChangePort);      
    addCmdLineOption(quietopts, 2,  SetQuiet); 
    addCmdLineOption(helpopts,  4,  DoHelp);
}

int SetQuiet (char * unused) {
  quiet = 1;
  return 0;
}

//user help
int DoHelp ( char * option) {
   printf(" Program options:\n\
   --help or -? or -h or ? \n\
      Shows this help\n\
      \n\
   -0 n or --off n\n\
      Turns off output n\n\
      Numbers are 1 based (start from 1)\n\
      \n\
   -1 n or --on n\n\
      Turn on output n\n\
      Numbers are 1 based(start from 1)\n\
      \n\
   --size n or -s n \n\
      Sets the number of output bits to n\n\
        By default all ports are off\n\
        The default size is 16\n\
      \n\
   --port n of -p n\n\
     Sets the io address of the parallel port to use\n\
     The address is in hexidecimal\n\
     The default address is 0x378\n\
     \n\
   --shutup or -q\n\
     Turns off anything but error messages\n\
     \n\
     \n\
     example:\n\
       porcon -s 96 -p 0x278 -0 2 -1 4\n\
      Sets the number of ports to 96, \n\
      the parallel port address to 0x278 \n\
      output 2 is turned off\n\
      output 4 is turned on\n\
      \n\
      \n\
      The settings are stored in /etc/superport.conf.\n\
      By default, the program does not change any of the settings\n\
      but will always re-transmitt the outputs.\n\
      \n\
      \n\
      so, if you now just run    porcon  the outputs will \n\
      stay the same OR resume their last state (if just powered up)\n\
   \n\
   ");
   justexit = 1;   
   return 0;
}

//resize array
int ChangePort ( char * option) {
   port = strtol( option, (char **)NULL, 16) ;   
   return 1;
}

//resize array
int ChangeBits ( char * option) {
  long newbits;
  
  newbits = strtol( option, (char **)NULL, 0) ; 
  AllocBits(newbits);  
  return 1;      
}

// wrapper for text input
int CBitOff ( char * option) {
  long bit;
  bit = strtol( option, (char **)NULL, 0) ;
  BitOff( bit-1 );  
  if (!quiet) printf("Will be Clearing bit: %ld\n", bit);
  return 1;  
}

// wrapper for text input
int CBitOn ( char * option) {
  long bit;
  bit = strtol( option, (char **)NULL, 0) ;
  BitOn( bit-1 );  
  if (!quiet) printf("Will be Setting bit:  %ld\n", bit);
  return 1;  
}


void BitOff ( long Bit) {  
 if ((Bit < bits) && (Bit >= 0)) {
   BitList[Bit] = 0;
 }
}

void BitOn ( long Bit) {
 if ((Bit < bits) && (Bit >= 0)) {
   BitList[Bit] = 1;
 }
}

void SendBit(int value) {       // send a bit to the shift registers

 if (value == 0) {              // if sending a 0...
  parallel_WriteData(ioport, 7);// 0            // put the data low, clock low
  parallel_WriteData(ioport, 5);// 2            // put the data low, clock high
  parallel_WriteData(ioport, 7);// 0            // put the data low, clock low

 } else {                       // if sending a 1...
  parallel_WriteData(ioport, 6);// 1            // put the data high, clock low
  parallel_WriteData(ioport, 4);// 3            // put the data high, clock high
  parallel_WriteData(ioport, 6);// 1            // put the data high, clock low
 }

}

void Strobe(void) {             // move the shift register contents to the outputs
 parallel_WriteData(ioport, 7); // 0            // data low, clock low, strobe low
 parallel_WriteData(ioport, 3); // 4            // data low, clock low, strobe high
 parallel_WriteData(ioport, 7); // 0            // data low, clock low, strobe low
}

